defmodule FileStream.Storage.HTTP do
  @moduledoc """
  A module to stream HTTP backed files.
  """

  @behaviour FileStream.Storage

  ################################
  # FileStream.Storage Callbacks
  ################################

  @impl FileStream.Storage
  @spec init(keyword) :: keyword()
  def init(_opts) do
    []
  end

  @impl FileStream.Storage
  def into(_stream) do
    raise "Not implemented"
  end

  @impl FileStream.Storage
  def reduce(stream, acc, fun) do
    Stream.resource(
      fn -> init_read(stream) end,
      &read_stream/1,
      &exit_stream/1
    ).(acc, fun)
  end

  ################################
  # Private API
  ################################

  defp init_read(stream) do
    uri = to_string(stream.uri)
    request(stream, :get, uri, '', async: :once)
  end

  defp request(stream, method, uri, body_or_stream, opts, attempt \\ 1) do
    retries = 5

    case :hackney.request(method, uri, [], body_or_stream, opts) do
      {:ok, ref} ->
        FileStream.put_meta(stream, :ref, ref)

      {:error, :closed} when attempt < retries ->
        request(stream, method, uri, body_or_stream, opts, attempt + 1)

      {:error, reason} ->
        reason = format_error(reason)
        raise FileStream.Error, uri: stream.uri, reason: reason
    end
  end

  defp read_stream(stream) do
    ref = FileStream.get_meta(stream, :ref)

    with {:error, reason} <- :hackney.stream_next(ref) do
      reason = format_error(reason)
      raise FileStream.Error, uri: stream.uri, reason: reason
    end

    receive do
      {:hackney_response, ^ref, {:status, 200, _}} ->
        {[], stream}

      {:hackney_response, ^ref, {:status, status, _}} ->
        reason = format_error({:status, status})
        raise FileStream.Error, uri: stream.uri, reason: reason

      {:hackney_response, ^ref, {:headers, _headers}} ->
        {[], stream}

      {:hackney_response, ^ref, chunk} when is_binary(chunk) ->
        {[chunk], stream}

      {:hackney_response, ^ref, :done} ->
        {:halt, stream}

      {:tcp_closed, _} ->
        reason = format_error(:tcp_closed)
        raise FileStream.Error, uri: stream.uri, reason: reason
    end
  end

  defp exit_stream(stream) do
    FileStream.reset_meta(stream)
  end

  # credo:disable-for-next-line
  defp format_error(error) do
    case error do
      :nxdomain -> "Domain does not exist."
      :req_not_found -> "Stream already complete."
      :tcp_closed -> "TCP closed."
      {:status, 404} -> "Does not exist."
      {:tls_alert, {:certificate_expired, _}} -> "SSL certificate expired."
      {:tls_alert, {:handshake_failure, _}} -> "SSL handhsake failure."
      {:tls_alert, {:bad_certificate, _}} -> "SSL bad certificate."
      {:tls_alert, {:unknown_ca, _}} -> "SSL unknown CA."
      {:tls_alert, {:revoked, _}} -> "SSL revoked."
    end
  end
end
